from series.errors import SeriesException


class ArchiverError(SeriesException):
    pass


class InvalidDBError(SeriesException):

    def __init__(self, specifics):
        text = 'Invalid database: {}'
        super().__init__(text.format(specifics))



import re
from datetime import datetime, date, timedelta

from sqlalchemy.sql.expression import or_, and_
from sqlalchemy.sql.elements import not_
from sqlalchemy.sql.functions import func

from golgi import configurable

from amino import List, Maybe, _, __, Try, Map, L

from series.get.model.release import ReleaseFactory, ReleaseMonitor, Release
from series.get.model.link import Link, Torrent
from series.db_facade import DbFacade, exclusive, commit
from series.util import datetime_to_unix


@configurable(get=['status_new_interval'])
class ReleasesFacade(DbFacade):

    @property
    def main_type(self):
        return ReleaseMonitor

    @property
    def name(self):
        return 'release'

    @property
    def monitors(self):
        return self.all_query

    def filter_release(self, *filters):
        return (self.monitors
                .join(Release)
                .filter(*filters)
                .order_by(ReleaseMonitor.id))

    def filter_by_release(self, **filters):
        return (self.monitors
                .join(Release)
                .filter_by(**filters)
                .order_by(ReleaseMonitor.id))

    def query_release(self, order=True, **filters):
        q = self._db.query(Release)
        if order:
            q = q.order_by(Release.id)
        return q.filter_by(**filters)

    def filter_by_metadata(self, series=None, season=None, episode=None):
        filters = List(
            Maybe(series) / (lambda a: Release.name == a),
            Maybe(season) / str / (lambda a: Release.season == a),
            Maybe(episode) / str / (lambda a: Release.episode == a),
        )
        return self.filter_release(*filters.join)

    @exclusive
    def find_by_metadata(self, series=None, season=None, episode=None):
        return self.filter_by_metadata(series, season, episode).first()

    def filter_episode_repr(self, regex):
        rex = re.compile(regex, re.I)
        matcher = lambda m: rex.search(str(m.release))
        return self.all.filter(matcher)

    def __getitem__(self, slice):
        return self.all[slice]

    def create(self, series, season, episode, airdate=None, search_name=None,
               **kw):
        fact = ReleaseFactory()
        enum = '{:0>2}x{:0>2}'.format(season, episode)
        title = '{}_{}'.format(series, enum)
        release = fact(title, series, '', '', enum=enum,
                       search_name=search_name)
        if airdate:
            release.airdate = airdate
        monitor = fact.monitor(release, [], [], **kw)
        self.add(monitor)
        return monitor

    def delete(self, series, season, episode):
        self._delete(self.find_by_metadata(series, season, episode))

    @exclusive
    def _delete(self, record):
        if record:
            self._db.session.delete_then_commit(record)

    @commit
    def _update(self, record, data):
        if record:
            for key, value in data.items():
                setattr(record, key, value)
        return record

    def update(self, series, season, episode, data):
        release = self.find_by_metadata(series, season, episode)
        return self._update(release, data)

    def update_by_id(self, _id, **data):
        release = self.find_by_id(_id)
        return self._update(release, data)

    def update_link(self, _id, **data):
        return self.link_by_id(_id) % L(self._update)(_, data)

    @property
    def pending_downloads(self):
        qualify = lambda r: not r.downloaded and not r.nuked
        return list(filter(qualify, self.all))

    def add_link_by_id(self, _id, url):
        release = self.find_by_id(_id)
        if release:
            self._add_links(release, List(url))
            return True

    @commit
    def _add_links(self, release, urls):
        release.add_links(urls)

    def add_torrent_by_id(self, _id, url):
        release = self.find_by_id(_id)
        if release:
            self._add_torrents(release, List(url))
            return True

    @commit
    def _add_torrents(self, release, urls):
        release.add_torrents(urls)

    def add_torrent(self, release, url):
        if release.has_torrent(url):
            text = 'Release {rel} already contains torrent "{link}"'
        else:
            self._add_torrents(release, List(url))
            text = 'Adding torrent "{link}" to release {rel}'
        self.log.info(text.format(link=url, rel=release.release))

    def add_links_from_feed_entry(self, release, feed_entry):
        new = [l for l in feed_entry.links if not release.has_url(l)]
        if new:
            self._add_links(release, new)
            text = 'Added new links to release {}: {}'
            self.log.info(text.format(release.release, ', '.join(new)))

    def add_link(self, release, url):
        if release.has_url(url):
            text = 'Release {rel} already contains link "{link}"'
        else:
            self._add_links(release, List(url))
            text = 'Adding link "{link}" to release {rel}'
        self.log.info(text.format(link=url, rel=release.release))

    @property  # type: ignore
    @exclusive
    def count(self):
        return self.monitors.count()

    @property  # type: ignore
    @exclusive
    def downloadable(self):
        return List.wrap(self.download_candidates.all()).filter(_.downloadable)

    @property
    def download_candidates(self):
        return self._db\
            .query(ReleaseMonitor)\
            .filter(not_(or_(ReleaseMonitor.downloaded, ReleaseMonitor.nuked,
                         ReleaseMonitor.archived, ReleaseMonitor.downloading)))

    @commit
    def add(self, release):
        self._db.session.add_then_commit(release)

    def add_season(self, name, season, count):
        for episode in range(1, count + 1):
            if not self._release_exists(name, season, episode):
                self.create(name, season, episode)

    def _release_exists(self, name, season, episode):
        return (self.query_release(name=name, season=season,
                                   episode=episode).count() > 0)

    def mark_episode_downloaded(self, monitor):
        self._mark_episode(monitor, 'downloaded')

    def mark_episode_archived(self, monitor):
        self._mark_episode(monitor, 'archived')

    @commit
    def _mark_episode(self, monitor, attr):
        r = self.query_release(id=monitor.release_id).first()
        if r:
            matches = self.filter_by_release(name=r.name, season=r.season,
                                             episode=r.episode)
        for m in matches:
            setattr(m, attr, True)

    def latest_for_season(self, name, season):
        q = (
            self._db.query(func.max(Release.episode))
            .filter_by(name=name, season=season)
            .join(ReleaseMonitor)
        )
        return Maybe(*q.first())

    def one(self, name, season, episode):
        rel = self.query_release(name=name, season=season,
                                 episode=episode).first()
        if rel:
            return self.filter_by(release_id=rel.id).first()

    @exclusive
    def reset_torrent(self, id):
        release = self.filter_by(id=id).first()
        if release:
            self._reset_torrent(release)

    @commit
    def _reset_torrent(self, release):
        for torrent in release.torrents.all():
            torrent.dead = True
        release.downloaded = False
        release.archived = False
        release.last_torrent_search_stamp = 0

    def http_link_by_id(self, id):
        return Maybe(self._db.query(Link).filter_by(id=id).first())

    def torrent_by_id(self, id):
        return Maybe(self._db.query(Torrent).filter_by(id=id).first())

    def link_by_id(self, id):
        return self.torrent_by_id(id).o(self.http_link_by_id(id))

    @commit
    def torrent_cached(self, id):
        return self.update_link(id, cached=True, caching=False)

    def fail_link(self, id):
        self.link_by_id(id).foreach(__.check_failed())

    @commit
    def set_airdate(self, id, target):
        return (self.by_id(id).zip(self._parse_date(target))
                .map2(self._set_airdate))

    def _set_airdate(self, release, stamp):
        release.release.airdate_stamp = datetime_to_unix(stamp)

    def _parse_date(self, target):
        error = lambda x: '\'{}\' does not match [YYYY-]MM-DD'.format(target)
        this_year = date.today().year
        if re.match('\d\d?-\d\d?', target):
            target = '{}-{}'.format(this_year, target)
        return Try(datetime.strptime, target, '%Y-%m-%d').lmap(error)

    @property  # type: ignore
    @exclusive
    def status(self):
        thresh = datetime_to_unix(
            datetime.now() - timedelta(days=self._status_new_interval))
        done = List.wrap((
            self.filter_by(downloaded=True)
            .filter(ReleaseMonitor.download_date_stamp > thresh)
        ))
        not_cached = List.wrap((
            self.monitors
            .filter(~ReleaseMonitor.downloaded)
            .filter(~ReleaseMonitor.torrents.any(
                and_(Torrent.cached, ~Torrent.dead)))
        ))
        caching, search = not_cached.split(__.caching_f())
        downloading = List.wrap(
            self.monitors.filter(ReleaseMonitor.downloading))
        failed = List.wrap((
            self.monitors
            .filter(~ReleaseMonitor.downloaded)
            .filter(ReleaseMonitor.failed_downloads > 0)
        ))
        return Map(
            done=done / _.done_info,
            caching=caching / _.release.info,
            search=search / _.release.info,
            downloading=downloading / _.release.info,
            failed=failed / _.release.info,
        )

__all__ = ('ReleasesFacade',)



import shutil, os
from os import path

from series.logging import series_logger
from tek.user_input import YesNo
from tek.tools import listdir_abs, ymap

from series import episode_metadata
from series.errors import SeriesException

log = series_logger('rename')

class EpisodeEnumerationError(SeriesException):
    def __init__(self, filename):
        text = "Couldn't find enumeration: %s" % filename
        TException.__init__(self, text)

def make_identifier(se):
    return 'x'.join(str(s).zfill(2) for s in se)

def rename(filename, series_name=None, season_name=None):
    ext = filename.rsplit('.', 1)[-1]
    _series_name, _season_name, _number = episode_metadata(filename)
    name = series_name or _series_name
    season = season_name or _season_name
    identifier = make_identifier((season, _number))
    return '%s_%s.%s' % (name, identifier, ext.lower())

def move_path(filepath, series_name=None, force=False):
    filename = path.basename(filepath)
    new_name = rename(filename, series_name)
    if not new_name:
        return None
        log.warning('%s cannot be renamed' % filename)
    else:
        return new_name

def move(filepath, series_name=None, force=False):
    if not path.exists(filepath):
        raise IOError('File not found: ' + filepath)
    dir = path.dirname(filepath)
    new_name = move_path(filepath, series_name, force)
    new_path = os.path.join(dir, new_name)
    exists = path.exists(new_path)
    prompt = '%s "%s"?' % (('Overwrite' if exists else 'Move to'), new_name)
    if exists and not force:
        log.info('File exists: ' + new_name)
    elif YesNo(prompt).read():
        shutil.move(filepath, new_path)

def move_description(job, exists, maxlen):
    s = job[0].ljust(maxlen) + '  =>  ' + job[1]
    if exists:
        s += ' (exists)'
    return s

def fix_all(dir='.', series_name=None):
    jobs = []
    files_abs = list(filter(path.isfile, listdir_abs(dir)))
    subs = path.join(dir, 'sub')
    if path.isdir(subs):
        files_abs.extend(list(filter(path.isfile, listdir_abs(subs))))
    files = list(map(os.path.basename, files_abs))
    dests = ymap(move_path, files, series_name)
    jobs = list(zip(files, dests, files_abs))
    jobs = [s_d_ig for s_d_ig in jobs if s_d_ig[1] is not None and s_d_ig[0] != s_d_ig[1]]
    if jobs:
        maxlen = max(len(j[0]) for j in jobs)
        exists = lambda f: os.path.isfile(os.path.join(dir, f))
        strings = [move_description(job, exists(job[1]), maxlen) for job in
                   jobs]
        log.info(strings)
        if YesNo(['Execute?']).read():
            for source, dest, source_abs in jobs:
                base = os.path.dirname(source_abs)
                shutil.move(source_abs, os.path.join(base, dest))

__all__ = ('find_episode_enumeration', 'make_series_name',
           'make_identifier', 'rename', 'move', 'fix_all',
           'EpisodeEnumerationError')



from golgi.errors import Error


class MissingMetadata(Error):
    pass


class SeriesException(Error):
    pass


class InvalidDBError(SeriesException):
    def __init__(self, specifics):
        text = 'Invalid database: {}'
        super().__init__(text.format(specifics))


class SeriesDException(SeriesException):
    pass



import re
import itertools

import requests

import lxml

from amino.lazy import lazy
from amino import List, LazyList, _, L

from tek_utils.sharehoster.torrent import SearchResultFactory
from tek_utils.sharehoster.kickass import NoResultsError

from series.get import ReleaseMonitor
from series.logging import Logging


class SearchQuery:

    def __init__(self, monitor: ReleaseMonitor, res: str) -> None:
        self.monitor = monitor
        self.release = self.monitor.release
        self.res = res

    @property
    def _enum(self):
        return 's{:0>2}e{:0>2}'.format(self.release.season,
                                       self.release.episode)

    @property
    def _name(self):
        return (
            self.release.effective_search_name
            .replace('_', ' ')
            .replace('\'', '')
        )

    @property
    def query(self):
        return '{} {} {}'.format(
            self._name,
            self._enum,
            self.res,
        )

    @property
    def valid(self):
        return True

    @property
    def search_string(self):
        return self.release.search_string_with_res(self.res, False)

    @lazy
    def search_re(self):
        return re.compile(self.search_string, re.I)

    @property
    def desc(self):
        return 'torrent {} {}'.format(self.release, self.res)


class DateQuery(SearchQuery):

    @property
    def _enum(self):
        return self.release.airdate.strftime('%Y-%m-%d')

    @property
    def valid(self):
        return self.release.has_airdate

    @property
    def search_string(self):
        return self.release.search_string_with_res(self.res, True)


class SearchResults:

    def __init__(self, monitor, query, results, min_seeders) -> None:
        self.monitor = monitor
        self.query = query
        self.results = results
        self.min_seeders = min_seeders

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.results)

    @property
    def choose(self):
        matcher = self.query.search_re
        valid = (
            self.results
            .filter(lambda a: a.seeders is not None)
            .filter(_.seeders > self.min_seeders)
        )
        return (
            valid
            .filter(lambda a: matcher.search(a.title))
            .filter(_.magnet_link)
            .filter_not(lambda a: self.monitor.contains_torrent(a.magnet_link))
            .head
        )


class TorrentSearch(Logging):

    def __init__(self, requester, search_engine, min_seeders) -> None:
        self.requester = requester
        self.search_engine = search_engine
        self.min_seeders = min_seeders
        self._limit = 10

    @lazy
    def _search(self):
        return (self._search_tpb if self.search_engine == 'piratebay' else
                self._search_kickass)

    def _search_tpb(self, query):
        from tek_utils.sharehoster import piratebay
        bay = piratebay.Search(query)
        return bay.run[:self._limit] / SearchResultFactory.from_tpb

    def _search_kickass(self, query):
        from tek_utils.sharehoster import kickass
        search = kickass.Search(query).order(kickass.ORDER.SEED,
                                             kickass.ORDER.DESC)
        return List.wrap(SearchResultFactory.from_kickass(res) for res in
                         itertools.islice(search, self._limit))

    def _queries(self, monitor):
        q = lambda r: List(SearchQuery(monitor, r), DateQuery(monitor, r))
        return monitor.resolutions // q

    def search(self, monitor):
        return (
            LazyList(self._queries(monitor))
            .filter(_.valid)
            .apzip(self._safe_search)
            .map2(L(SearchResults)(monitor, _, _, self.min_seeders))
        )

    def _safe_search(self, query):
        self.log.debug(
            'Search {} for {}: {}'.format(query.desc, self.requester,
                                          query.query))
        try:
            return List.wrap(self._search(query.query))
        except NoResultsError as e:
            self.log.debug('Error searching for torrent: {}'.format(e))
        except requests.RequestException as e:
            self.log.warning(
                'Connection failure in {} search'.format(self.search_engine))
        except lxml.etree.XMLSyntaxError as e:
            self.log.warning('Parse error in kickass results: {}'.format(e))
        return List()

__all__ = ('TorrentSearch',)



"""remove link polymorphism

Revision ID: d363aeb3c4de
Revises: a49797d68d62
Create Date: 2016-09-25 01:01:56.204651

"""

# revision identifiers, used by Alembic.
revision = 'd363aeb3c4de'
down_revision = 'a49797d68d62'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('http_link')
    op.drop_table('torrent_link')
    op.add_column('link', sa.Column('checking', sa.Boolean(), nullable=True))
    op.add_column('link', sa.Column('dead', sa.Boolean(), nullable=True))
    op.add_column('link', sa.Column('multipart', sa.Boolean(), nullable=True))
    op.add_column('link', sa.Column('reason', sa.String(), nullable=True))
    op.add_column('link', sa.Column('size', sa.Integer(), nullable=True))
    op.add_column('link', sa.Column('status', sa.Integer(), nullable=True))
    op.add_column('link', sa.Column('time_checked', sa.Integer(), nullable=True))
    op.drop_index('ix_link_real_type', table_name='link')
    op.add_column('release_monitor', sa.Column('download_date_stamp', sa.Integer(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release_monitor', sa.Column('torrent_cached', sa.BOOLEAN(), nullable=True))
    op.add_column('link', sa.Column('real_type', sa.VARCHAR(length=24), nullable=False))
    op.create_index('ix_link_real_type', 'link', ['real_type'], unique=False)
    op.create_table('torrent_link',
    sa.Column('created_at', sa.DATETIME(), nullable=True),
    sa.Column('updated_at', sa.DATETIME(), nullable=True),
    sa.Column('id', sa.INTEGER(), nullable=False),
    sa.Column('cached', sa.BOOLEAN(), nullable=True),
    sa.Column('dead', sa.BOOLEAN(), nullable=True),
    sa.ForeignKeyConstraint(['id'], ['link.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('http_link',
    sa.Column('status', sa.INTEGER(), nullable=True),
    sa.Column('dead', sa.BOOLEAN(), nullable=True),
    sa.Column('multipart', sa.BOOLEAN(), nullable=True),
    sa.Column('time_checked', sa.INTEGER(), nullable=True),
    sa.Column('checking', sa.BOOLEAN(), nullable=True),
    sa.Column('size', sa.INTEGER(), nullable=True),
    sa.Column('reason', sa.VARCHAR(), nullable=True),
    sa.Column('created_at', sa.DATETIME(), nullable=True),
    sa.Column('updated_at', sa.DATETIME(), nullable=True),
    sa.Column('id', sa.INTEGER(), nullable=False),
    sa.ForeignKeyConstraint(['id'], ['link.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    ### end Alembic commands ###
